@@ -11,6 +11,19 @@
 #include <caml/callback.h>
 #include <caml/unixsupport.h>
 #include <caml/socketaddr.h>
+#include <caml/version.h>
+#if OCAML_VERSION < 41300
+#define CAML_INTERNALS /* was needed from OCaml 4.06 to 4.12 */
+#endif
+#include <caml/osdeps.h>
+
+#if OCAML_VERSION_MAJOR < 5
+#define caml_unix_error_of_code unix_error_of_code
+#define caml_uerror uerror
+#define caml_win32_maperr win32_maperr
+#define caml_win32_alloc_handle win_alloc_handle
+#define caml_win32_alloc_socket win_alloc_socket
+#endif
 
 //#define D(x) x
 #define D(x) while(0){}
@@ -76,8 +89,8 @@ static void invoke_completion_callback
   err = Val_long(0);
   if (errCode != NO_ERROR) {
     len = -1;
-    win32_maperr (errCode);
-    err = unix_error_of_code(errno);
+    caml_win32_maperr(errCode);
+    err = caml_unix_error_of_code(errno);
   }
   name = caml_copy_string(action_name[action]);
   D(printf("Action %s completed: id %ld -> len %ld / err %d (errCode %ld)\n",
@@ -154,8 +167,8 @@ static HANDLE get_helper_thread (value threads, int kind) {
 
   h = CreateThread (NULL, 0, helper_thread, NULL, 0, NULL);
   if (h == NULL) {
-    win32_maperr (GetLastError ());
-    uerror("createHelperThread", Nothing);
+    caml_win32_maperr(GetLastError());
+    caml_uerror("createHelperThread", Nothing);
   }
   Field(threads, kind) = (value) h;
   return h;
@@ -238,7 +251,7 @@ static void thread_io
   ioInfo * info = GlobalAlloc(GPTR, sizeof(ioInfo));
   if (info == NULL) {
     errno = ENOMEM;
-    uerror(action_name[action], Nothing);
+    caml_uerror(action_name[action], Nothing);
   }
 
   info->action = action;
@@ -267,7 +280,7 @@ static void overlapped_action(long action, long id,
   completionData * d = GlobalAlloc(GPTR, sizeof(completionData));
   if (d == NULL) {
     errno = ENOMEM;
-    uerror(action_name[action], Nothing);
+    caml_uerror(action_name[action], Nothing);
   }
   d->id = id;
   d->action = action;
@@ -281,10 +294,10 @@ static void overlapped_action(long action, long id,
   if (!res) {
     err = GetLastError ();
     if (err != ERROR_IO_PENDING) {
-      win32_maperr (err);
+      caml_win32_maperr(err);
   D(printf("Action %s failed: id %ld -> err %d (errCode %ld)\n",
            action_name[action], id, errno, err));
-      uerror("ReadFileEx", Nothing);
+      caml_uerror("ReadFileEx", Nothing);
     }
   }
 }
@@ -345,46 +358,47 @@ static void after_connect (SOCKET s) {
 static HANDLE events[MAXIMUM_WAIT_OBJECTS];
 //static OVERLAPPED oData[MAXIMUM_WAIT_OBJECTS];
 
-CAMLprim value win_register_wait (value socket, value kind, value idx) {
-  CAMLparam3(socket, kind, idx);
-  long i = Long_val(idx);
+CAMLprim value win_register_wait (value socket, value kind) {
+  CAMLparam2(socket, kind);
+  HANDLE h;
   long mask;
 
-  D(printf("Register: i %ld, kind %ld\n", Long_val(i), Long_val(kind)));
-  events[i] = CreateEvent(NULL, TRUE, FALSE, NULL);
+  D(printf("Register: %lx, kind %ld\n", (long)(Socket_val(socket)), Long_val(kind)));
+  h = CreateEvent(NULL, TRUE, FALSE, NULL);
   mask = (Long_val(kind) == 0) ? FD_CONNECT : FD_ACCEPT;
-  if (WSAEventSelect(Socket_val(socket), events[i], mask) == SOCKET_ERROR) {
-    win32_maperr(WSAGetLastError ());
-    uerror("WSAEventSelect", Nothing);
+  if (WSAEventSelect(Socket_val(socket), h, mask) == SOCKET_ERROR) {
+    caml_win32_maperr(WSAGetLastError());
+    caml_uerror("WSAEventSelect", Nothing);
   }
 
-  CAMLreturn (Val_unit);
+  CAMLreturn(caml_win32_alloc_handle(h));
 }
 
-CAMLprim value win_check_connection (value socket, value kind, value idx) {
-  CAMLparam3 (socket, kind, idx);
+CAMLprim value win_check_connection (value socket, value kind, value h) {
+  CAMLparam3 (socket, kind, h);
   WSANETWORKEVENTS evs;
-  int res, err, i = Long_val(idx);
+  int res, err;
 
-  D(printf("Check connection... %d\n", i));
+  D(printf("Check connection... socket = %lx; h = %lx\n",
+              (long)(Socket_val(socket)), Handle_val(h)));
   if (WSAEnumNetworkEvents(Socket_val(socket), NULL, &evs)) {
-    win32_maperr(WSAGetLastError ());
-    uerror("WSAEnumNetworkEvents", Nothing);
+    caml_win32_maperr(WSAGetLastError());
+    caml_uerror("WSAEnumNetworkEvents", Nothing);
   }
   if (WSAEventSelect(Socket_val(socket), NULL, 0) == SOCKET_ERROR) {
-    win32_maperr(WSAGetLastError ());
-    uerror("WSAEventSelect", Nothing);
+    caml_win32_maperr(WSAGetLastError());
+    caml_uerror("WSAEventSelect", Nothing);
   }
-  if (!CloseHandle(events[i])) {
-    win32_maperr(GetLastError ());
-    uerror("CloseHandle", Nothing);
+  if (!CloseHandle(Handle_val(h))) {
+    caml_win32_maperr(GetLastError());
+    caml_uerror("CloseHandle", Nothing);
   }
   err =
     evs.iErrorCode[(Long_val(kind) == 0) ? FD_CONNECT_BIT : FD_ACCEPT_BIT];
   D(printf("Check connection: %ld, err %d\n", evs.lNetworkEvents, err));
   if (err != 0) {
-    win32_maperr(err);
-    uerror("check_connection", Nothing);
+    caml_win32_maperr(err);
+    caml_uerror("check_connection", Nothing);
   }
   CAMLreturn (Val_unit);
 }
@@ -422,14 +436,16 @@ CAMLprim value init_lwt (value callback) {
   CAMLreturn (Val_long (MAXIMUM_WAIT_OBJECTS));
 }
 
-CAMLprim value win_wait (value timeout, value event_count) {
-  CAMLparam2(timeout, event_count);
+CAMLprim value win_wait (value timeout, value event_list) {
+  CAMLparam2(timeout, event_list);
   DWORD t, t2;
   DWORD res;
-  long ret, n = Long_val(event_count);
+  long ret, n;
   t = Long_val(timeout);
   if (t < 0) t = INFINITE;
   t2 = (compN > 0) ? 0 : t;
+  for (n = 0; event_list != Val_emptylist; event_list = Field(event_list, 1))
+    events[n++] = Handle_val(Field(event_list, 0));
   D(printf("Waiting: %ld events, timeout %ldms -> %ldms\n", n, t, t2));
   res =
     (n > 0) ?
@@ -449,8 +465,8 @@ CAMLprim value win_wait (value timeout, value event_count) {
   case WAIT_FAILED:
     D(printf("Wait failed\n"));
     ret = 0;
-    win32_maperr (GetLastError ());
-    uerror("WaitForMultipleObjectsEx", Nothing);
+    caml_win32_maperr(GetLastError());
+    caml_uerror("WaitForMultipleObjectsEx", Nothing);
     break;
   default:
     ret = res;
@@ -483,8 +499,8 @@ value win_pipe(long readMode, long writeMode) {
      1, UNIX_BUFFER_SIZE, UNIX_BUFFER_SIZE, 0, &attr);
 
   if (readh == INVALID_HANDLE_VALUE) {
-    win32_maperr(GetLastError());
-    uerror("CreateNamedPipe", Nothing);
+    caml_win32_maperr(GetLastError());
+    caml_uerror("CreateNamedPipe", Nothing);
     return FALSE;
   }
 
@@ -494,17 +510,17 @@ value win_pipe(long readMode, long writeMode) {
      FILE_ATTRIBUTE_NORMAL | writeMode, NULL);
 
   if (writeh == INVALID_HANDLE_VALUE) {
-    win32_maperr(GetLastError());
+    caml_win32_maperr(GetLastError());
     CloseHandle(readh);
-    uerror("CreateFile", Nothing);
+    caml_uerror("CreateFile", Nothing);
     return FALSE;
   }
 
-  readfd = win_alloc_handle(readh);
-  writefd = win_alloc_handle(writeh);
+  readfd = caml_win32_alloc_handle(readh);
+  writefd = caml_win32_alloc_handle(writeh);
   res = caml_alloc_small(2, 0);
-  Store_field(res, 0, readfd);
-  Store_field(res, 1, writefd);
+  Field(res, 0) = readfd;
+  Field(res, 1) = writefd;
   CAMLreturn (res);
 }
 
@@ -519,7 +535,7 @@ CAMLprim value win_pipe_out (value unit) {
 }
 
 static int socket_domain_table[] = {
-  PF_UNIX, PF_INET
+  PF_UNIX, PF_INET, PF_INET6
 };
 
 static int socket_type_table[] = {
@@ -536,10 +552,10 @@ CAMLprim value win_socket (value domain, value type, value proto) {
                 NULL, 0, WSA_FLAG_OVERLAPPED);
   D(printf("Created socket %lx\n", (long)s));
   if (s == INVALID_SOCKET) {
-    win32_maperr(WSAGetLastError ());
-    uerror("WSASocket", Nothing);
+    caml_win32_maperr(WSAGetLastError());
+    caml_uerror("WSASocket", Nothing);
   }
-  CAMLreturn(win_alloc_socket(s));
+  CAMLreturn(caml_win32_alloc_socket(s));
 }
 
 /*
@@ -610,7 +626,7 @@ CAMLprim value win_readdirtorychanges
   completionData * d = GlobalAlloc(GPTR, sizeof(completionData));
   if (d == NULL) {
     errno = ENOMEM;
-    uerror(action_name[action], Nothing);
+    caml_uerror(action_name[action], Nothing);
   }
   d->id = id;
   d->action = action;
@@ -624,10 +640,10 @@ CAMLprim value win_readdirtorychanges
   if (!res) {
     err = GetLastError ();
     if (err != ERROR_IO_PENDING) {
-      win32_maperr (err);
+      caml_win32_maperr(err);
   D(printf("Action %s failed: id %ld -> err %d (errCode %ld)\n",
            action_name[action], id, errno, err));
-      uerror("ReadDirectoryChangesW", Nothing);
+      caml_uerror("ReadDirectoryChangesW", Nothing);
     }
   }
   CAMLreturn (Val_unit);
@@ -638,14 +654,16 @@ CAMLprim value win_parse_directory_changes (value buf_val) {
   CAMLlocal4(lst, tmp, elt, filename);
   char * pos = Array_data(buf_val, 0);
   FILE_NOTIFY_INFORMATION * entry;
+  wchar_t *namebuf;
 
   lst = Val_long(0);
   while (1) {
     entry = (FILE_NOTIFY_INFORMATION *)pos;
+    namebuf = calloc(entry->FileNameLength + 2, 1);
+    memmove(namebuf, entry->FileName, entry->FileNameLength);
     elt = caml_alloc_tuple(2);
-    filename = caml_alloc_string(entry->FileNameLength);
-    memmove((char *)String_val(filename), entry->FileName, entry->FileNameLength);
-    Store_field (elt, 0, filename);
+    Store_field (elt, 0, caml_copy_string_of_utf16(namebuf));
+    free(namebuf);
     Store_field (elt, 1, Val_long(entry->Action - 1));
     tmp = caml_alloc_tuple(2);
     Store_field (tmp, 0, elt);
@@ -657,31 +675,34 @@ CAMLprim value win_parse_directory_changes (value buf_val) {
   CAMLreturn(lst);
 }
 
-CAMLprim value win_open_directory (value path, value wpath) {
-  CAMLparam2 (path, wpath);
+CAMLprim value win_open_directory (value path) {
+  CAMLparam1 (path);
   HANDLE h;
-  h = CreateFileW((LPCWSTR) String_val(wpath),
+  wchar_t *wpath = caml_stat_strdup_to_utf16(String_val(path));
+
+  h = CreateFileW(wpath,
                   FILE_LIST_DIRECTORY,
                   FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                   NULL,
                   OPEN_EXISTING,
                   FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED,
                   NULL);
+  caml_stat_free(wpath);
   if (h == INVALID_HANDLE_VALUE) {
-    win32_maperr (GetLastError ());
-    uerror("open", path);
+    caml_win32_maperr(GetLastError());
+    caml_uerror("open", path);
   }
-  CAMLreturn(win_alloc_handle(h));
+  CAMLreturn(caml_win32_alloc_handle(h));
 }
 
-value copy_wstring(LPCWSTR s);
-
 CAMLprim value win_long_path_name(value path) {
   CAMLparam1(path);
+  wchar_t *wpath = caml_stat_strdup_to_utf16(String_val(path));
   wchar_t lbuf[32768] = L"";
   DWORD res;
 
-  res = GetLongPathNameW((LPCWSTR) String_val(path), lbuf, 32768);
+  res = GetLongPathNameW(wpath, lbuf, 32768);
+  caml_stat_free(wpath);
 
-  CAMLreturn(res == 0 || res > 32767 ? path : copy_wstring(lbuf));
+  CAMLreturn(res == 0 || res > 32767 ? path : caml_copy_string_of_utf16(lbuf));
 }
