@@ -136,7 +136,7 @@ let encodeException m kind e =
     Unix.Unix_error(err,fnname,param) ->
       let s =   "Error in " ^ m ^ ":\n"
               ^ (Unix.error_message err)
-              ^ " [" ^ fnname ^ "(" ^ param ^ ")]%s" ^
+              ^ " [" ^ fnname ^ "(" ^ param ^ ")]" ^
               (match err with
                  Unix.EUNKNOWNERR n -> Format.sprintf " (code %d)" n
                | _                  -> "")
@@ -328,6 +328,23 @@ let percentageOfTotal current total =
 
 let percent2string p = Printf.sprintf "%3d%%" (truncate (max 0. (min 100. p)))
 
+let gib = 1073741824.
+let mib = 1048576.
+let kib = 1024.
+let bytes2string v =
+  if v > 1_048_051_711L then
+    Printf.sprintf "%.2f GiB" (Int64.to_float v /. gib)
+  else if v > 104_805_171L then
+    Printf.sprintf "%.0f MiB" (Int64.to_float v /. mib)
+  else if v > 1_023_487L then
+    Printf.sprintf "%.1f MiB" (Int64.to_float v /. mib)
+  else if v > 102_348L then
+    Printf.sprintf "%.0f KiB" (Int64.to_float v /. kib)
+  else if v > 999L then
+    Printf.sprintf "%.1f KiB" (Int64.to_float v /. kib)
+  else
+    Printf.sprintf "%Ld B" v
+
 let extractValueFromOption = function
     None -> raise (Fatal "extractValueFromOption failed")
   | Some(v) -> v
@@ -461,7 +478,7 @@ let padto n s = s ^ (String.make (max 0 (n - String.length s)) ' ')
 (*              Building pathnames in the user's home dir                    *)
 (*****************************************************************************)
 
-let homeDirStr = 
+let homeDir () =
     (if (osType = `Unix) || isCygwin then
        safeGetenv "HOME"
      else if osType = `Win32 then
@@ -482,10 +499,7 @@ let homeDirStr =
      else
        assert false (* osType can't be anything else *))
 
-let homeDir () =
-  System.fspathFromString homeDirStr
-
-let fileInHomeDir n = System.fspathConcat (homeDir ()) n
+let fileInHomeDir n = Filename.concat (homeDir ()) n
 
 (*****************************************************************************)
 (*                       .unison dir                                         *)
@@ -497,7 +511,7 @@ let isMacOSX = isMacOSXPred ()
 
 let unisonDir =
   try
-    System.fspathFromString (System.getenv "UNISON")
+    System.getenv "UNISON"
   with Not_found ->
     let genericName =
       fileInHomeDir (Printf.sprintf ".%s" ProjectInfo.myName) in
@@ -506,11 +520,9 @@ let unisonDir =
     else
       genericName
 
-let unisonDirStr = System.fspathToString unisonDir
-
-let fileInUnisonDir str = System.fspathConcat unisonDir str
+let fileInUnisonDir str = Filename.concat unisonDir str
 
 let fileMaybeRelToUnisonDir n =
   if Filename.is_relative n
   then fileInUnisonDir n
-  else System.fspathFromString n
+  else n
