@@ -25,7 +25,8 @@ let debug = Trace.debug "ui"
 let dumbtty =
   Prefs.createBool "dumbtty"
     (try System.getenv "EMACS" <> "" with Not_found -> false)
-    "!do not change terminal settings in text UI"
+    ~category:(`Advanced `CLI)
+    "do not change terminal settings in text UI"
     ("When set to \\verb|true|, this flag makes the text mode user "
      ^ "interface avoid trying to change any of the terminal settings.  "
      ^ "(Normally, Unison puts the terminal in `raw mode', so that it can "
@@ -40,7 +41,9 @@ let dumbtty =
      ^ "interface.")
 
 let silent =
-  Prefs.createBool "silent" false "print nothing except error messages"
+  Prefs.createBool "silent" false
+    ~category:(`Basic `Syncprocess_CLI)
+    "print nothing except error messages"
     ("When this preference is set to {\\tt true}, the textual user "
      ^ "interface will print nothing at all, except in the case of errors.  "
      ^ "Setting \\texttt{silent} to true automatically sets the "
@@ -87,8 +90,9 @@ let setupTerminal() =
       restoreTerminal ()
 
 let colorMode =
-  Prefs.createBoolWithDefault "color" ~local:true
-    "!use color output for text UI (true/false/default)"
+  Prefs.createBoolWithDefault "color"
+    ~category:(`Advanced `CLI) ~local:true
+    "use color output for text UI (true/false/default)"
     ("When set to {\\tt true}, this flag enables color output in "
      ^ "text mode user interface. When set to {\\tt false}, all "
      ^ "color output is disabled. Default is to enable color if "
@@ -461,7 +465,7 @@ let interact prilist rilist =
           begin match !Prefs.profileName with None -> assert false |
             Some(n) ->
               display ("  To un-ignore, edit "
-                       ^ System.fspathToPrintString (Prefs.profilePathname n)
+                       ^ Prefs.profilePathname n
                        ^ " and restart " ^ Uutil.myName ^ "\n") end;
           let nukeIgnoredRis =
             Safelist.filter (fun ri -> not (Globals.shouldIgnore ri.path1)) in
@@ -797,12 +801,16 @@ let doTransport reconItemList =
          (fun s item -> Uutil.Filesize.add item.bytesToTransfer s)
          Uutil.Filesize.zero items)
   in
+  let totalBytesToTransferStr = Util.bytes2string
+    (Uutil.Filesize.toInt64 !totalBytesToTransfer) in
   let t0 = Unix.gettimeofday () in
   let calcProgress i bytes dbg =
     let i = Uutil.File.toLine i in
     let item = items.(i) in
     item.bytesTransferred <- Uutil.Filesize.add item.bytesTransferred bytes;
     totalBytesTransferred := Uutil.Filesize.add !totalBytesTransferred bytes;
+    let totalBytesTransferredStr = Util.bytes2string
+      (Uutil.Filesize.toInt64 !totalBytesTransferred) in
     let v =
       (Uutil.Filesize.percentageOfTotalSize
          !totalBytesTransferred !totalBytesToTransfer)
@@ -810,12 +818,18 @@ let doTransport reconItemList =
     let t1 = Unix.gettimeofday () in
     let remTime =
       if v <= 0. then "--:--"
-      else if v >= 100. then "00:00"
+      else if v >= 100. then "00:00:00"
       else
         let t = truncate ((t1 -. t0) *. (100. -. v) /. v +. 0.5) in
-        Format.sprintf "%02d:%02d" (t / 60) (t mod 60)
+        let u = t mod 3600 in
+        let h = t / 3600 in
+        let m = u / 60 in
+        let sec = u mod 60 in
+        Format.sprintf "%02d:%02d:%02d" h m sec
     in
-    t1, Format.sprintf "%s  %s ETA" (Util.percent2string v) remTime
+    let stat = Format.sprintf "%s  (%s of %s)  %s ETA" (Util.percent2string v)
+      totalBytesTransferredStr totalBytesToTransferStr remTime in
+    t1, stat
   in
   let tlog = ref t0 in
   let showProgress i bytes dbg =
@@ -888,10 +902,7 @@ let doTransport reconItemList =
 
 let setWarnPrinterForInitialization()=
   Util.warnPrinter :=
-     Some(fun s ->
-            alwaysDisplay "Error: ";
-            alwaysDisplay (s^"\n");
-            exit Uicommon.fatalExit)
+    Some (fun s -> alwaysDisplay ("Warning: " ^ s ^ "\n\n"))
 
 let setWarnPrinter() =
   Util.warnPrinter :=
@@ -1096,7 +1107,7 @@ let checkForDangerousPath dangerousPaths =
     end
   end
 
-let synchronizeOnce ?wantWatcher ?skipRecentFiles pathsOpt =
+let synchronizeOnce ?wantWatcher pathsOpt =
   let showStatus path =
     if path = "" then Util.set_infos "" else
     let max_len = 70 in
@@ -1157,7 +1168,7 @@ module PathMap = Map.Make (Path)
 
 let waitForChangesRoot: Common.root -> unit -> unit Lwt.t =
   Remote.registerRootCmd
-    "waitForChanges"
+    "waitForChanges" Umarshal.unit Umarshal.unit
     (fun (fspath, _) -> Fswatchold.wait (Update.archiveHash fspath))
 
 let waitForChanges t =
@@ -1180,7 +1191,7 @@ let synchronizePathsFromFilesystemWatcher () =
         delayInfo ([], [])
     in
     let (exitStatus, failedPaths) =
-      synchronizeOnce ~wantWatcher:() ~skipRecentFiles:()
+      synchronizeOnce ~wantWatcher:true
         (if isStart then None else Some (readyPaths, delayedPaths))
     in
     (* After a failure, we retry at once, then use an exponential backoff *)
@@ -1210,10 +1221,10 @@ let synchronizePathsFromFilesystemWatcher () =
 (* ----------------- Repetition ---------------- *)
 
 let synchronizeUntilNoFailures repeatMode =
+  let wantWatcher = repeatMode in
   let rec loop triesLeft pathsOpt =
     let (exitStatus, failedPaths) =
-      synchronizeOnce
-        ?wantWatcher:(if repeatMode then Some () else None) pathsOpt in
+      synchronizeOnce ~wantWatcher pathsOpt in
     if failedPaths <> [] && triesLeft <> 0 then begin
       loop (triesLeft - 1) (Some (failedPaths, []))
     end else begin
@@ -1249,7 +1260,9 @@ let rec synchronizeUntilDone () =
 
 let profmgrPrefName = "i"
 let profmgrPref =
-  Prefs.createBool profmgrPrefName false ~local:true
+  Prefs.createBool profmgrPrefName false
+    ~category:(`Basic `CLI)
+    ~cli_only:true
     "interactive profile mode (text UI); command-line only"
     ("Provide this preference in the command line arguments to enable "
      ^ "interactive profile manager in the text user interface. Currently "
@@ -1284,18 +1297,26 @@ let addProfileKeys list default =
   in
   addKey 0 [] list
 
+let scanProfiles () =
+  let wp = !Util.warnPrinter in
+  (* Replace warn printer with something that doesn't quit
+     the UI just for errors in random scanned profiles. *)
+  Util.warnPrinter := Some (fun s -> alwaysDisplay ("Warning: " ^ s ^ "\n\n"));
+  let () = Uicommon.scanProfiles () in
+  Util.warnPrinter := wp
+
 let getProfile default =
   let cmdArgs = Prefs.scanCmdLine Uicommon.shortUsageMsg in
-  Uicommon.scanProfiles ();
   if Util.StringMap.mem Uicommon.runTestsPrefName cmdArgs ||
     not (Util.StringMap.mem profmgrPrefName cmdArgs) then
     Some default
   else
+  let () = scanProfiles () in
   if (List.length !Uicommon.profilesAndRoots) > 10 then begin
     Trace.log (Format.sprintf "You have too many profiles in %s \
                 for interactive selection. Please specify profile \
                 or roots on command line.\n"
-                (System.fspathToPrintString Util.unisonDir));
+                Util.unisonDir);
     Trace.log "The profile names are:\n";
     Safelist.iter (fun (p, _) -> Trace.log (Format.sprintf "  %s\n" p))
       !Uicommon.profilesAndRoots;
@@ -1374,26 +1395,43 @@ let rec start interface =
   begin try
     (* Just to make sure something is there... *)
     setWarnPrinterForInitialization();
-    Uicommon.uiInit
-      ~reportError:
-      (fun s -> Util.msg "%s%s\n\n%s\n" Uicommon.shortUsageMsg profmgrUsageMsg s; exit 1)
-      ~tryAgainOrQuit:
-      (fun s -> Util.msg "%s" Uicommon.shortUsageMsg; exit 1)
+    let errorOut s =
+      Util.msg "%s%s%s\n" Uicommon.shortUsageMsg profmgrUsageMsg s;
+      exit 1
+    in
+    let profileName = match Uicommon.uiInitClRootsAndProfile () with
+      | Error s -> errorOut ("\n\n" ^ s)
+      | Ok None ->
+          let profile = getProfile "default" in
+          let () = restoreTerminal () in
+          begin
+            match profile with
+            | None -> exit 0
+            | Some x -> x
+          end
+      | Ok (Some s) -> s
+    in
+    Uicommon.initPrefs
+      ~profileName
       ~displayWaitMessage:
       (fun () -> setWarnPrinter();
                  if Prefs.read silent then Prefs.set Trace.terse true;
                  if not (Prefs.read silent)
                  then Util.msg "%s\n" (Uicommon.contactingServerMsg()))
-      ~getProfile:
-      (fun () -> let prof = getProfile "default" in restoreTerminal(); prof)
-      ~getFirstRoot:
-      (fun () -> Util.msg "%s%s\n" Uicommon.shortUsageMsg profmgrUsageMsg; exit 1)
-      ~getSecondRoot:
-      (fun () -> Util.msg "%s%s\n" Uicommon.shortUsageMsg profmgrUsageMsg; exit 1)
+      ~promptForRoots:
+      (fun () -> errorOut "")
       ~termInteract:
       None
       ();
 
+    if Prefs.read Uicommon.testServer then exit 0;
+
+    (* Run unit tests if requested *)
+    if Prefs.read Uicommon.runtests then begin
+      !Uicommon.testFunction ();
+      exit 0
+    end;
+
     (* Some preference settings imply others... *)
     if Prefs.read silent then begin
       Prefs.set Globals.batch true;
